# Implementation of approximate SDD-TMPC (state-dependent dynamic tube model predictive control) for a nonlinear problem

## Project description 

The code in this repository is used to implement and test an approximation of a new algorithm called SDD-TMPC. It is a control algorithm that sacrifices some optimality (much less than tube MPC), but provides robust solutions (more about this can be found in [1]). The original controller presented in [1] proved to work and showed good performance, but solving the optimality problem took too long to implement it for a real robot. To solve this problem, we used a spiking neural network to approximate SDD-TMPC. 

First, we generated data using Matlab's main script. It is a function that takes a seed as input and randomly samples the state of the robot in an environment as it works. The environment is a circular environment with 1 meter radius and a wall around it. Then it tries to solve the optimization problem, i.e. find an optimal input such that a cost function is minimized, a robot does not collide with the wall, and the terminal constraints are satisfied. Once the optimization problem is solved, a new file is created and the sampled data is saved with the additional data. 

It is possible that no solution was found. In such a case, a vector full of zeros will be saved. It is worth noting that it takes a lot of time to generate each data sample. According to our research, it is important to generate the data as close to the global optimum as possible, because outliers make training the network much more difficult. Therefore, we decided to spend a lot of time generating each data sample.


All generated data were saved in two files "final_data.csv" and "final_data.xlsx". The former contains the state (only 2 columns, there is no saved direction because it is assumed that the cooridanate frame is rotated so that the direction is always 0), optimal input (2 columns - linear and angular velocity), nominal input (2 columns - linear and angular velocity) and nominal error (3 columns - 2-dimensional position and direction). The nominal velocity and nominal error can be used to calculate the optimal input (see [2] for details), i.e. to check if there is no problem with the data. The file "final_data.xlsx" also contains an additional column with seeds to make the data reproducible. 

The next step is to use "network_traning.ipynb" to train the networks. After running this file, the files "linear_final.pt" and "angular_final.pt" are created. These files contain the trained spiking neural networks that can be used instead of SDD-TMPC. The "network_tester.ipynb" file is used to compare the behavior between SDD-TMPC and its approximation. After execution, figures are generated.

Finally, to test the behavior of the system, we created a ROS2 package to control a wheeled robot.



## How to use

Run main.m function in Matlab where the input argument is a seed -> generate data for training (as mat file)

Run jupyter notebook "Network_traning.ipynb" -> train networks using previously generated data. Note that the previously generated data must be saved as CSV files instead of mat files. 

Run jupyter notebook "Network_tester.ipynb" -> Compare the trained networks with SDD-TMPC and generate images.

To use the ROS2 package, the package must be copied into the ROS2 workspace. (See the online ROS2 tutorials for more information). Then call SNN_controller with the ros2 run command. It is important to note that SNN_controller is designed to work with cameras from our lab. To use the software in your application, you would need to change the names of the themes and/or the coordinate system (in our lab we use a coordinate system where y points up).

## Dependencies

Matlab's optimization toolbox (to use interior-point algorithm) - https://nl.mathworks.com/products/optimization.html

Matlab's global optimization toolbox (to use particle swarm) - https://nl.mathworks.com/products/global-optimization.html

MPT3 tooblox - To do some polyhedron calculations

Matlab's engine - Used to run Matlab's function in Python. Version 9.14.3. Note: It is important to have Matlab's engine that supports both Python and Matlab.

Numpy - Version 1.26.1

PyTorch - Used to implement connections between layers of the network - 2.2.0

SNNtorch - Used to implement spiking neural networks - Version 0.7.0

Matplotlib - Used to create plots - Version 3.5.1

TQDM - Used to create progress bars - Version 4.66.1

ROS2 Humble - Used to use the ROS2 package

## Known problems

There may be a problem loading the neural network from files. Normally it should be enough to import the original class, but for some reason this does not always work. In such a case it is important to add two lines: "import main" and "setattr(__main__, "$", $)" where $ has to be replaced with the name of the class containing the neural network definition. 

Using the MPT3 toolbox from Python requires reloading the toolbox. To do this, all toolbox files have been copied to the repository.

While using ROS2 package to control the robot, the robot is heavily affected by the delays which can lead to instability. 

The current ROS2 package was written for a general application to show the behavior of the controller. Since it was not written with neuromorphic hardware in mind, the controller is not even close to its maximum computational efficiency.

The software only works with SNN-Torch 0.7.0. Both older and newer versions made it impossible to load the neural network.

## License

The software is distributed under the MIT license:


Copyright (c) 2024 Filip Surma

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE

## References


1. Surma, F. Jamshidnejad, A.(2023). State-Dependent Dynamic Tube MPC: A Novel tube MPC Method with a Fuzzy Model of Disturbances [Manuscript submitted for publication]. Aerospace Engineering, Tu Delft.

2. Surma, F. Jamshidnejad, A.(2024). Approximate SDD-TMPC with Spiking Neural Networks: An Application to Wheeled Robots [Manuscript submitted for publication]. Aerospace Engineering, Tu Delft.
