function c = SDDTMPC_nonlcon(u,x,p,model_function,number_of_inputs,k,E,G,terminal_constant,horizon,distur_function,Ts,world_size)
%{
This function checks if hard constraints of SDD-TMPC are satisfited in the chosen horizon. 

:vector u: traejctory of inputs
:vector x: initial state vector state
:float p: robot's body's radius
:function model_function: model used to predict future states
:unsigned int number_of_inputs: Number of inputs of the controlled system
:matrix k: state feedback's gain matrix
:Matrix/vector E,G: state constraints such that E*x<=G
:float terminal_constant: This constant is used to define the terminal set tube. The
tube is a box defined as {|x[1]|<terminal_constant,|x[2]|<terminal_constant,|x[3]|<pi}
:uint horizon: horizon of the MPC
:function distur_function: state-dependent model of distrbunce
:float Ts: discretization step
:float world_size: A world is cricle with radius equla to world_size

:float^2 return: If c<0, then the constraints are satifited. Ceq is always 0 and is added for compatiliblity with other Matlab solvers.
%}

c=horizon;
z = x+u((end-2):end)';
error = x-z;
x_error_set = ones(2,1)*error(1);
y_error_set = ones(2,1)*error(2);
theta_error_set = ones(2,1)*error(3);

for i=1:horizon
    v=u((number_of_inputs*(i-1)+1):number_of_inputs*i)';
    
    [min_sin,max_sin,min_cos,max_cos] = get_border_sin_cos(z(3),theta_error_set);
    d = distur_function(theta_error_set,v,max(abs([min_sin,max_sin])),max(abs([min_cos,max_cos])))*Ts;
    xy_error_verticies = get_xy_verticies(x_error_set,y_error_set);
    disc_system_min = c2d(ss([-k(1,1),0,0;0,-k(2,2),0;-(1/p)*[min_sin,min_cos,v(1)]],[],[],[]),Ts);
    disc_system_max = c2d(ss([-k(1,1),0,0;0,-k(2,2),0;-(1/p)*[max_sin,max_cos,v(1)]],[],[],[]),Ts);
    scaled_error_x = [min([disc_system_min.A(3,1)*(x_error_set);disc_system_max.A(3,1)*(x_error_set)]),...
        max([disc_system_min.A(3,1)*(x_error_set);disc_system_max.A(3,1)*(x_error_set)])]';
    scaled_error_y = [min([disc_system_min.A(3,2)*(y_error_set);disc_system_max.A(3,2)*(y_error_set)]),...
        max([disc_system_min.A(3,2)*(y_error_set);disc_system_max.A(3,2)*(y_error_set)])]';
    x_error_set = x_error_set*disc_system_min.A(1,1)+d(1,:)';
    y_error_set = y_error_set*disc_system_min.A(2,2)+d(2,:)';
    xy_error_verticies_transformed = (rot(z(3),p)^(-1))*(-k)*xy_error_verticies+v;
    U_set_scaling=get_U_set_scaling(theta_error_set); 
    theta_error_set = disc_system_min.A(3,3)*theta_error_set+scaled_error_x+scaled_error_y+d(3,:)';
    
    
    
    robot_tube = get_tube_of_full_robot(theta_error_set+z(3),p,xy_error_verticies')+z(1:2);
    z = model_function(z,v);
    robot_ver = robot_tube.V;
    
    failure_flag= false;
    for j=1:4
        if ~all(E*xy_error_verticies_transformed(:,j)<U_set_scaling*G)
            failure_flag=true;
            break
        end
    end

    for j=1:length(robot_ver)
        if norm(robot_ver(j,:))>world_size
            failure_flag=true;
            break
        end
    end


    if failure_flag
        break
    end
    c=c-1;

end

c = c+max(max(abs(z(1:2)))-terminal_constant,0);


end