function [inputs,fval,exitflag,message] = SDDTMPC_solver(cost_func,warm_start,x0,model_function,K,p,...
    lb,ub,E,G,tube_constant,terminal_constant,horizon,distur_function,max_iter,Ts,world_size)
%{
This function finds a minimum of a chosen cost function under constraints of SDD-TMPC.  

:function cont_func: function to be minimized
:matrix warm_start: an initial swarm
:vector x0: state
:function model_function: model used to predict future states
:matrix K: state feedback's gain matrix
:float p: robot's body's radius
:vector^2 lb/ub: minimum/maximum inputs
:Matrix/vector E,G: state constraints such that E*x<=G
:float tube_constant: This constant is used to define the biggest tube. The
tube is a box defined as {|x[1]|<tube_constant,|x[2]|<tube_constant,|x[3]|<pi}
:float terminal_constant: This constant is used to define the terminal set tube. The
tube is a box defined as {|x[1]|<terminal_constant,|x[2]|<terminal_constant,|x[3]|<pi}
:uint horizon: horizon of the MPC
:function distur_function: state-dependent model of distrbunce
:int max_iter: the optimization will be stopped if number of iterations exceed this value.
:float Ts: discretization step
:float world_size: A world is cricle with radius equla to world_size
 
:return: 
	:vector inputs: optimal inputs
	:float fval: optimal cost
	:int exitflag: reason why optimization stopped (read https://nl.mathworks.com/help/gads/particleswarm.html to know more)
	:struct massege: addational information about optimization (read https://nl.mathworks.com/help/gads/particleswarm.html to know more)
	
%}
InitialSwarmMatrix=warm_start';
number_of_inputs = length(lb);
lb= [repmat(lb,[horizon,1]);-tube_constant*ones(2,1);-pi];
ub= [repmat(ub,[horizon,1]);tube_constant*ones(2,1);pi];
nvars = number_of_inputs*horizon+length(x0);


cost_func_tranposed = @(u) cost_func(u')+10^10*SDDTMPC_nonlcon(u,x0,...
    p,model_function,2,K,E,G,terminal_constant,horizon,distur_function,Ts,world_size);


options = optimoptions('particleswarm','SwarmSize',50,'MaxIterations',max_iter,...'InitialSwarmMatrix',warm_start',
    'Display','iter','DisplayInterval',10,'MaxStallIterations',max_iter, ...
    'InitialSwarmMatrix',InitialSwarmMatrix);

[inputs,fval,exitflag,message]=particleswarm(cost_func_tranposed,nvars,lb,ub,options);
inputs=inputs';

end

