import rclpy
from rclpy.node import Node
import math
import numpy as np

from geometry_msgs.msg import PoseStamped,Pose,Twist

def get_quaternion_from_euler(roll, pitch, yaw):
  """
  Convert an Euler angle to a quaternion.
   
  Input
    :param roll: The roll (rotation around x-axis) angle in radians.
    :param pitch: The pitch (rotation around y-axis) angle in radians.
    :param yaw: The yaw (rotation around z-axis) angle in radians.
 
  Output
    :return qx, qy, qz, qw: The orientation in quaternion [x,y,z,w] format
  """
  qx = np.sin(roll/2) * np.cos(pitch/2) * np.cos(yaw/2) - np.cos(roll/2) * np.sin(pitch/2) * np.sin(yaw/2)
  qy = np.cos(roll/2) * np.sin(pitch/2) * np.cos(yaw/2) + np.sin(roll/2) * np.cos(pitch/2) * np.sin(yaw/2)
  qz = np.cos(roll/2) * np.cos(pitch/2) * np.sin(yaw/2) - np.sin(roll/2) * np.sin(pitch/2) * np.cos(yaw/2)
  qw = np.cos(roll/2) * np.cos(pitch/2) * np.cos(yaw/2) + np.sin(roll/2) * np.sin(pitch/2) * np.sin(yaw/2)
 
  return [qx, qy, qz, qw]

class MinimalPublisher(Node):

    def __init__(self):
        super().__init__('minimal_publisher')
        self.publisher_ = self.create_publisher(PoseStamped, 'mocap', 10)
        self.timer_period = 0.001  # seconds
        self.timer = self.create_timer(self.timer_period, self.timer_callback)
        self.u=[0,0]
        self.x=[0,0,0.1]
        self.p=0.4731/2
        self.subscription = self.create_subscription(
            Twist,
            'cmd_vel',
            self.listener_callback,
            10)
        self.subscription  # prevent unused variable warning


    def timer_callback(self):
        msg = Pose()
        self.x[0]=self.x[0]+self.timer_period*(self.u[0]*math.cos(self.x[2]))
        self.x[1]=self.x[1]+self.timer_period*(self.u[0]*math.sin(self.x[2]))
        self.x[2]=self.x[2]+self.timer_period*self.u[1]
        msg.position.z= self.x[0]
        msg.position.x= self.x[1]
        #msg.orientation.z= self.x[2]
        qx, qy, qz, qw  =  get_quaternion_from_euler(0,0,self.x[2])
        msg.orientation.x = qx
        msg.orientation.y = qz
        msg.orientation.z = qy
        msg.orientation.w = qw
        stamped_msg=PoseStamped()
        stamped_msg.pose=msg
        self.publisher_.publish(stamped_msg)
        
    def listener_callback(self,msg):
        self.u =[msg.linear.x,msg.angular.z]
        




def main(args=None):
    rclpy.init(args=args)

    minimal_publisher = MinimalPublisher()

    rclpy.spin(minimal_publisher)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    minimal_publisher.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()

